#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <sys/stat.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "core.h"
#include "preferences.h"
#include "mdsgtkhelp.h"



extern GtkWidget *mainw;
extern GtkWidget *searchw;
extern GtkWidget *infow;
extern GtkWidget *picturew;
extern GtkWidget *notepad;
static xmameprefs *actualrc = NULL;



void notepad_updatestatus(void);


static void set_scale(GtkScale *scale, double v)

{
 gtk_adjustment_set_value(gtk_range_get_adjustment(GTK_RANGE(scale)), v);
}





static double get_scale(GtkScale *scale)

{
 return gtk_range_get_adjustment(GTK_RANGE(scale))->value;
}



static void set_entry(GtkEntry *entry, gchar *txt)

{
 if (txt == NULL) txt = "";
 gtk_entry_set_text(entry, txt);
}



static void set_gnome_number_entry(GnomeNumberEntry *n, double val)

{
 gchar s[256];

 sprintf(s, "%2.2f", val);
 set_entry(GTK_ENTRY(gnome_number_entry_gtk_entry(n)), s);
}




static void get_entry(GtkEntry *entry, gchar **txt)

{
 g_free(*txt);
 *txt = g_strdup(gtk_entry_get_text(entry));
}



static GSList *get_clist(GtkCList *l)

{
 GSList *l1 = NULL;
 gchar *row[1];

 while (gtk_clist_get_text(l, 0, 0, row)) {;
    l1 = g_slist_append(l1, g_strdup(row[0]));
	 gtk_clist_remove(l, 0);
 }

 return (l1);
}




static void set_clist(GtkCList *l, GSList *l1)

{
 gchar *row[1];

 gtk_clist_freeze(l);
 gtk_clist_clear(l);
 while (l1 != NULL) {
	 row[0] = l1->data;
	 gtk_clist_append(l, row);
	 l1 = l1->next;
 }
 gtk_clist_thaw(l);
}




static void set_mapkeys(GtkCList *list, GSList *mapkeys)

{
 gchar *row[2];
 gchar *s, *s1, s2[256];

 gtk_entry_set_text(GTK_ENTRY(lookup_widget(GTK_WIDGET(list), "w_mapkey_lentry")), "");
 gtk_entry_set_text(GTK_ENTRY(lookup_widget(GTK_WIDGET(list), "w_mapkey_rentry")), "");
 gtk_clist_clear(list);
 while (mapkeys) {
	 s = mapkeys->data;
	 s1 = s2;
	 while (*s != '\0' && *s != ',') *s1++ = *s++;
	 *s1 = '\0';
	 row[0] = g_strdup(s2);
	 if (*s != '\0') s++;
	 s1 = s2;
	 while (*s != '\0') *s1++ = *s++;
	 *s1 = '\0';
	 row[1] = g_strdup(s2);
	 gtk_clist_append(list, row);
	 g_free(row[0]);
	 g_free(row[1]);
	 mapkeys = mapkeys->next;
 }
}



static GSList *get_mapkeys(GtkCList *l)

{
 GSList *l1 = NULL;
 gchar *row[1];
 gchar *s;

 while (gtk_clist_get_text(l, 0, 0, row)) {;
	 s = g_strdup(row[0]);
    gtk_clist_get_text(l, 0, 1, row);
    l1 = g_slist_append(l1, g_strconcat(s, ",", row[0], NULL));
	 g_free(s);
	 gtk_clist_remove(l, 0);
 }

 return (l1);
}




static GtkWidget *give_widget(GtkWidget *w, gchar *name)

{
 gchar s[256];
 gchar *ss;
 int i;

 strcpy(s, "w_");
 for (i = 2, ss = name; *ss != '\0'; i++, ss++)
	 s[i] = (*ss == '-') ? '_' : *ss;
 s[i] = '\0';

 return lookup_widget(w, s);
}




static void set_sensitivity(GtkWidget *w, gchar *name, gboolean go_down)

{
 xmameprefrule *r, *r1;
 int i, val = 0;
 gboolean enable, flag;
 GtkWidget *widg;
 gchar *ss;

 r1 = xmameprefrule_find(name);
 widg = give_widget(w, name);
 flag = r1->valid;
 if (flag && strcmp(name, "heightscale") == 0) {
    if (r1->depend) set_sensitivity(w, r1->depend, FALSE);
    flag = !GTK_TOGGLE_BUTTON(lookup_widget(w, "linkxy"))->active;
 }
 if (!flag) gtk_widget_set_sensitive(widg, flag);
 if (r1->widgtype == XMAMEPREF_WTYPE_CLIST) {
    gtk_widget_set_sensitive(give_widget(w,
   		                    ss = g_strconcat(name, "_add", NULL)),
	                          flag);
    g_free(ss);
    gtk_widget_set_sensitive(give_widget(w,
                             ss = g_strconcat(name, "_del", NULL)),
                             flag);
    g_free(ss);
    gtk_widget_set_sensitive(give_widget(w,
	                          ss = g_strconcat(name, "_clear", NULL)),
                             flag);
    g_free(ss);
 }
 for (i = 0; (r = xmameprefrule_get(i)); i++) {
	 if (r->valid && r->depend && r->widgtype != XMAMEPREF_WTYPE_NONE) {
		 if (strcmp(r->depend, name) == 0) {
          switch (r1->widgtype) {
          case XMAMEPREF_WTYPE_TOGGLE :
             val = GTK_TOGGLE_BUTTON(widg)->active;
             break;
          case XMAMEPREF_WTYPE_COMBO :
             val = gtk_combo_get_choice(GTK_COMBO(widg));
             break;
          case XMAMEPREF_WTYPE_OPMENU :
             val = gtk_option_menu_get_history(GTK_OPTION_MENU(widg));
             break;
          default : 
				 val = 0;
             fprintf(stderr, "Option : %s, Opt.type : %c, Widg.type : %c\n",
                     r1->name, r1->type, r1->widgtype);
             gnomame_error("Widget dependence not implemented",
                           GNOMAME_WARNING);
             break;
          }
          enable = r->negative ? (val != r->condvalue) : (val == r->condvalue);
		    gtk_widget_set_sensitive(give_widget(w, r->name), flag && enable);
		    if (go_down) set_sensitivity(w, r->name, TRUE);
       }
	 }
 }
}


 
 
static void update_xmamepropw(xmameprefs *prefs,
		                        GtkWidget *widget,
										gboolean flag)

{
 xmamepref *p;
 xmameprefrule *r;
 GSList *l;
 GList  *l1;
 GtkWidget *w;
 gchar s[256];
 int i;

 strcpy(s, "w_");
 for (i = 0; (r = xmameprefrule_get(i)); i++) {
	 if (r->widgtype != XMAMEPREF_WTYPE_NONE) {
	    w = give_widget(widget, r->name);
	    if (w) {
			 if (flag)
	          set_sensitivity(w, r->name, TRUE);
			 else if (r->valid) {
				 p = xmameprefs_find(prefs, r->name);
   		    if (strcmp(p->r->name, "heightscale") == 0)
   	          if (GTK_TOGGLE_BUTTON(lookup_widget(widget, "linkxy"))->active)
   		          p->val.f = xmameprefs_find(prefs, "widthscale")->val.f;
             switch (p->r->widgtype) {
             case XMAMEPREF_WTYPE_TOGGLE :
                p->val.b = GTK_TOGGLE_BUTTON(w)->active;
                break;
             case XMAMEPREF_WTYPE_SPIN :
                if (p->r->type == XMAMEPREF_TYPE_INT)
   	             p->val.i = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(w));
                else
	                p->val.f = gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(w));
                break;
             case XMAMEPREF_WTYPE_SCALE :
                if (p->r->type == XMAMEPREF_TYPE_INT)
      	          p->val.i = (int) get_scale(GTK_SCALE(w));
                else
	                p->val.f = get_scale(GTK_SCALE(w));
                break;
             case XMAMEPREF_WTYPE_ENTRY :
                get_entry(GTK_ENTRY(w), &p->val.s);
                break;
             case XMAMEPREF_WTYPE_CLIST :
                g_slist_foreach(p->val.l, (GFunc) g_free, NULL);
                g_slist_free(p->val.l);
   	          if (strcmp(p->r->name, "mapkey") == 0)
   		          p->val.l = get_mapkeys(GTK_CLIST(w));
                else
                   p->val.l = get_clist(GTK_CLIST(w));
                break;
             case XMAMEPREF_WTYPE_COMBO :
					 if (p->r->type == XMAMEPREF_TYPE_STRING) {
                   if (strcmp(p->r->name, "sound-mixer-plugin") == 0) {
							 g_free(p->val.s);
							 p->val.s = g_strdup(gnomame_plugin_code(gnomame_mixerplugins,
									                          gtk_combo_get_choice(GTK_COMBO(w))));
						 } else if (strcmp(p->r->name, "dsp-plugin") == 0) {
							 g_free(p->val.s);
							 p->val.s = g_strdup(gnomame_plugin_code(gnomame_dspplugins,
									                          gtk_combo_get_choice(GTK_COMBO(w))));
                   }
					 } else
	                p->val.i = gtk_combo_get_choice(GTK_COMBO(w));
                break;
             case XMAMEPREF_WTYPE_OPMENU :
	             p->val.i = gtk_option_menu_get_history(GTK_OPTION_MENU(w));
					 if (strcmp(p->r->name, "bpp") == 0) p->val.i *= 8;
                break;
             case XMAMEPREF_WTYPE_GNUMBR :
                p->val.f = gnome_number_entry_get_number((GnomeNumberEntry *) w);
                break;
             }
	       }
       }
    }
 }
 if (flag) {
	 if ((r = xmameprefrule_find("defaultgame")) && r->valid) {
       w = lookup_widget(widget, "defaultgameb");
       l1 = gnomame_give_availables();
       gtk_combo_set_popdown_strings(GTK_COMBO(w), l1);
       g_list_free(l1);
    }
	 if ((r = xmameprefrule_find("sound-mixer-plugin")) && r->valid) {
       w = give_widget(widget, r->name);
       l1 = gnomame_get_plugin_descriptions(gnomame_mixerplugins);
       gtk_combo_set_popdown_strings(GTK_COMBO(w), l1);
       g_list_free(l1);
    }
	 if ((r = xmameprefrule_find("dsp-plugin")) && r->valid) {
       w = give_widget(widget, r->name);
       l1 = gnomame_get_plugin_descriptions(gnomame_dspplugins);
       gtk_combo_set_popdown_strings(GTK_COMBO(w), l1);
       g_list_free(l1);
    }
    for (l = prefs->prefs; l; l = l->next) {
       p = (xmamepref *) l->data;
	    if (p->r->valid && p->r->widgtype != XMAMEPREF_WTYPE_NONE) {
          if ((w = give_widget(widget, p->r->name))) {
             switch (p->r->widgtype) {
             case XMAMEPREF_WTYPE_TOGGLE :
                gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(w), p->val.b);
                break;
             case XMAMEPREF_WTYPE_SPIN :
                gtk_spin_button_set_value(GTK_SPIN_BUTTON(w),
                       (p->r->type == XMAMEPREF_TYPE_INT) ? p->val.i : p->val.f);
                break;
             case XMAMEPREF_WTYPE_SCALE :
                set_scale(GTK_SCALE(w),
                       (p->r->type == XMAMEPREF_TYPE_INT) ? p->val.i : p->val.f);
                break;
             case XMAMEPREF_WTYPE_ENTRY :
                set_entry(GTK_ENTRY(w), p->val.s);
                break;
             case XMAMEPREF_WTYPE_CLIST :
		          if (strcmp(p->r->name, "mapkey") == 0)
			          set_mapkeys(GTK_CLIST(w), p->val.l);
		          else
			          set_clist(GTK_CLIST(w), p->val.l);
                break;
             case XMAMEPREF_WTYPE_COMBO :
					 if (p->r->type == XMAMEPREF_TYPE_STRING) {
						 if (strcmp(p->r->name, "sound-mixer-plugin") == 0) {
						    gtk_combo_set_choice(GTK_COMBO(w),
								    gnomame_plugin_code_to_num(gnomame_mixerplugins, p->val.s));
						 } else if (strcmp(p->r->name, "dsp-plugin") == 0) {
						    gtk_combo_set_choice(GTK_COMBO(w),
								    gnomame_plugin_code_to_num(gnomame_dspplugins, p->val.s));
                   }
					 } else {
			          if (strcmp(p->r->name, "bpp") == 0)
	                   gtk_combo_set_choice(GTK_COMBO(w),
									                       p->val.i / 8);
					    else
                      gtk_combo_set_choice(GTK_COMBO(w), p->val.i);
                }
                break;
             case XMAMEPREF_WTYPE_OPMENU :
			       if (strcmp(p->r->name, "bpp") == 0)
	                gtk_option_menu_set_history(GTK_OPTION_MENU(w),
						                       p->val.i / 8);
				    else
                   gtk_option_menu_set_history(GTK_OPTION_MENU(w), p->val.i);
                break;
             case XMAMEPREF_WTYPE_GNUMBR :
                set_gnome_number_entry((GnomeNumberEntry *) w, p->val.f);
                break;
             }
          }
       }
    }
 }
}


void
on_xmamepropw_show                     (GtkWidget       *widget,
                                        gpointer         user_data)
{
 gchar *s;

 if (actualrc) xmameprefs_free(actualrc);
 if (gnomame_selectedgame) {
    s = g_strconcat("Gnomame : ", gnomame_selectedgame->description, " settings", NULL);
    actualrc = xmameprefs_read(gnomame_selectedgame);
 } else {
    s = g_strconcat("Gnomame : General xmame settings", NULL);
    actualrc = xmameprefs_read(NULL);
 }
 gtk_window_set_title(GTK_WINDOW(widget), s);
 g_free(s);
 update_xmamepropw(actualrc, widget, TRUE);
}




void
on_okb_clicked                         (GtkButton       *button,
                                        gpointer         user_data)
{
 update_xmamepropw(actualrc, GTK_WIDGET(button), FALSE);
 xmameprefs_write(actualrc);
 if (actualrc->g) gnomame_refresh_game(actualrc->g);
 gtk_widget_destroy(lookup_widget(GTK_WIDGET(button), "xmamepropw"));
}



gboolean
on_mainw_delete_event                  (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 gnomame_exit();
 gtk_exit(0); /* Not called gtk_main_quit() because when loading games, we
					  must exit immediately without any event processing */

 return FALSE;
}


void
on_gnoprefb_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkWidget *w = create_gnopropw();

 gtk_window_set_modal(GTK_WINDOW(w), TRUE);
 gtk_widget_show(w);
}


void
on_searchb_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_widget_show(searchw);
}


void
on_aboutb_clicked                       (GtkButton       *button,
                                         gpointer         user_data)

{
 const gchar *authors[] = { "Denis Manente", NULL };
 GtkWidget *aboutw;

 aboutw = gnome_about_new ("GnoMAME", VERSION,
						   _("GNU General Public License"),
						   authors,
						   _("A Gnome/GTK+ xmame frontend"),
		                   "gnomame/gnomame.xpm");
 gtk_object_set_data (GTK_OBJECT (aboutw), "aboutw", aboutw);
 gtk_window_set_modal(GTK_WINDOW(aboutw), TRUE);
 gtk_widget_show(aboutw);
 gtk_signal_connect(GTK_OBJECT (aboutw), "delete_event",
                    GTK_SIGNAL_FUNC(gtk_widget_destroy),
                    (gpointer) aboutw);
}



static void open_xmamepropw(void)

{
 extern GtkWidget *confirmoptions;

 if (gnomame_selectedgame == NULL ||
	  gnomame_selectedgame->state > GNOMAME_WARNING) {
	 gnomame_unselect_all();
    gtk_widget_show(create_xmamepropw());
 } else
    gtk_menu_popup(GTK_MENU(confirmoptions), NULL, NULL, NULL, NULL, 0, 0);
}



void
on_xmameprefb_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 open_xmamepropw();
}



void
on_preview_realize                     (GtkWidget       *widget,
                                        gpointer         user_data)
{
 extern GtkWidget *preview;
 
 preview = widget;
}


void
on_infobox_realize                        (GtkWidget       *widget,
                                           gpointer         user_data)
{
 extern GtkText *info;
 
 info = GTK_TEXT(widget);
}



void
on_previewb_toggled                    (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 if (togglebutton->active)
    gtk_widget_show(picturew);
 else
	 gtk_widget_hide(picturew);
 gnomame_show_preview(togglebutton, gnomame_selectedgame, preview);
}


void
on_infob_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 if (togglebutton->active)
    gtk_widget_show(infow);
 else
	 gtk_widget_hide(infow);
 gnomame_show_info(togglebutton, gnomame_selectedgame, info);
}




gboolean
on_infow_delete_event                  (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 GtkToggleButton *b = GTK_TOGGLE_BUTTON(lookup_widget(mainw, "infob"));

 gtk_toggle_button_set_active(b, FALSE);

 return TRUE;
}




gboolean
on_picturew_delete_event               (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 GtkToggleButton *b = GTK_TOGGLE_BUTTON(lookup_widget(mainw, "previewb"));

 gtk_toggle_button_set_active(b, FALSE);

 return TRUE;
}



static void gnopropw_setget(GtkWidget *w, gboolean setget)

{
 GtkToggleButton *rb1, *rb2, *rb3;
 GtkToggleButton *rb11, *rb12, *rb13, *rb14;
 GtkToggleButton *showinfo, *showpreview, *confirmplay;
 GtkToggleButton *showcol[4];
 GtkEntry *xmamepath, *previewdir;
 gchar s[30];
 gint i;

 previewdir = GTK_ENTRY(lookup_widget(w, "previewdir"));
 xmamepath = GTK_ENTRY(lookup_widget(w, "xmamepath"));
 showinfo = GTK_TOGGLE_BUTTON(lookup_widget(w, "showinfo"));
 showpreview = GTK_TOGGLE_BUTTON(lookup_widget(w, "showpreview"));
 confirmplay = GTK_TOGGLE_BUTTON(lookup_widget(w, "confirmplay"));
 rb1 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb1"));
 rb2 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb2"));
 rb3 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb3"));
 rb11 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb11"));
 rb12 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb12"));
 rb13 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb13"));
 rb14 = GTK_TOGGLE_BUTTON(lookup_widget(w, "rb14"));
 for (i = 0; i < 4; i++) {
	 sprintf(s, "showcol_%d", i);
	 showcol[i] = GTK_TOGGLE_BUTTON(lookup_widget(w, s));
 }

 if (setget) {
    switch (gnomamerc->toolbartype) {
    case 1 : gtk_toggle_button_set_active(rb2, TRUE); break;
    case 2 : gtk_toggle_button_set_active(rb3, TRUE); break;
    default : gtk_toggle_button_set_active(rb1, TRUE); break;
    }
    switch (gnomamerc->status) {
    case 1 : gtk_toggle_button_set_active(rb12, TRUE); break;
    case 2 : gtk_toggle_button_set_active(rb13, TRUE); break;
    case 3 : gtk_toggle_button_set_active(rb14, TRUE); break;
    default : gtk_toggle_button_set_active(rb11, TRUE); break;
    }
    gtk_toggle_button_set_active(showinfo, gnomamerc->showinfo);
    gtk_toggle_button_set_active(showpreview, gnomamerc->showpreview);
    gtk_toggle_button_set_active(confirmplay, gnomamerc->confirmplay);
    gtk_entry_set_text(previewdir, gnomamerc->previewdir);
    gtk_entry_set_text(xmamepath, gnomamerc->xmamepath);
    for (i = 0; i < 4; i++)
       gtk_toggle_button_set_active(showcol[i], gnomamerc->showcol[i]);
 } else {
    if (rb2->active) gnomamerc->toolbartype = 1;
    else if (rb3->active) gnomamerc->toolbartype = 2;
    else gnomamerc->toolbartype = 0;
    if (rb12->active) gnomamerc->status = 1;
    else if (rb13->active) gnomamerc->status = 2;
    else if (rb14->active) gnomamerc->status = 3;
    else gnomamerc->status = 0;
    gnomamerc->showinfo = showinfo->active;
    gnomamerc->showpreview = showpreview->active;
    gnomamerc->confirmplay = confirmplay->active;
    g_free(gnomamerc->previewdir);
    g_free(gnomamerc->xmamepath);
    gnomamerc->previewdir = g_strdup(gtk_entry_get_text(previewdir));
    gnomamerc->xmamepath = g_strdup(gtk_entry_get_text(xmamepath));
    for (i = 0; i < 4; i++) gnomamerc->showcol[i] = showcol[i]->active;
 }
}



void
on_gnopropw_show                       (GtkWidget       *widget,
                                        gpointer         user_data)
{
 gnopropw_setget(widget, TRUE);
}


void
on_gnomameokb_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 gnopropw_setget(GTK_WIDGET(button), FALSE);
 gnomameprefs_write(gnomamerc);
 gtk_widget_destroy(GTK_WIDGET(user_data));
}




void
on_linkxy_toggled                      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 GtkSpinButton *hs, *ws;

 hs = GTK_SPIN_BUTTON(give_widget(GTK_WIDGET(togglebutton), "heightscale"));
 if (togglebutton->active) {
    ws = GTK_SPIN_BUTTON(give_widget(GTK_WIDGET(togglebutton), "widthscale"));
    gtk_spin_button_set_value(hs, gtk_spin_button_get_value_as_float(ws));
 }

 set_sensitivity(GTK_WIDGET(togglebutton), "heightscale", TRUE);
}


void
on_sound_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "sound", TRUE);
}





static gboolean is_dir(gchar *filename)

{
 struct stat buf;

 if (!stat(filename, &buf)) return S_ISDIR(buf.st_mode);
 return (FALSE);
}



static void filesel_click(GtkWidget *w, GtkCList *rompaths)

{
 GtkFileSelection *fs = GTK_FILE_SELECTION(gtk_widget_get_toplevel(w));
 gchar *row[1], *s;

 if (fs->ok_button == w) {
	 if (is_dir(s = gtk_file_selection_get_filename(fs))) {
       row[0] = g_strdup(s);
       gtk_clist_append(rompaths, row);
    } else
		 return;
 }
 gtk_widget_destroy(GTK_WIDGET(fs));
}




void
on_addrompath_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkFileSelection *filesel;

 filesel = GTK_FILE_SELECTION(gtk_file_selection_new("New ROM path"));
 gtk_signal_connect(GTK_OBJECT(filesel),
		              "destroy", (GtkSignalFunc) gtk_widget_destroy,
						  filesel);
 gtk_signal_connect(GTK_OBJECT(filesel->ok_button),
			           "clicked", (GtkSignalFunc) filesel_click,
						  user_data);
 gtk_signal_connect(GTK_OBJECT(filesel->cancel_button),
	                 "clicked", (GtkSignalFunc) filesel_click,
						  user_data);
 gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
 gtk_widget_show(GTK_WIDGET(filesel));
}


void
on_deleterompath_clicked               (GtkButton       *button,
                                        gpointer        user_data)
{
 GtkCList *l = GTK_CLIST(user_data);

 gtk_clist_freeze(l);
 while (l->selection != NULL)
	 gtk_clist_remove(l, GPOINTER_TO_INT(l->selection->data));
 gtk_clist_thaw(l);
}




void
on_clearrompaths_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_clist_clear(GTK_CLIST(user_data));
}



void
on_tweak_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "tweak", TRUE);
}



void
on_widthscale_changed                  (GtkEditable     *editable,
                                        gpointer         user_data)
{
 GtkToggleButton *lxy;
 GtkSpinButton *hs;
 
 lxy = GTK_TOGGLE_BUTTON(lookup_widget(GTK_WIDGET(editable), "linkxy"));
 if (lxy->active) {
	 hs = GTK_SPIN_BUTTON(give_widget(GTK_WIDGET(editable), "heightscale"));
    gtk_spin_button_set_value(hs,
			 gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(editable)));
 }
}





void
on_rb1_toggled                         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 GtkToolbar *tb = GTK_TOOLBAR(lookup_widget(mainw, "toolbar"));

 gtk_toolbar_set_style(tb, GTK_TOOLBAR_ICONS);
}


void
on_rb2_toggled                         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 GtkToolbar *tb = GTK_TOOLBAR(lookup_widget(mainw, "toolbar"));

 gtk_toolbar_set_style(tb, GTK_TOOLBAR_BOTH);
}


void
on_rb3_toggled                         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 GtkToolbar *tb = GTK_TOOLBAR(lookup_widget(mainw, "toolbar"));

 gtk_toolbar_set_style(tb, GTK_TOOLBAR_TEXT);
}


void
on_mainw_show                          (GtkWidget       *widget,
                                        gpointer         user_data)
{
 GtkToggleButton *b;
 GtkToolbar *tb = GTK_TOOLBAR(lookup_widget(widget, "toolbar"));

 switch (gnomamerc->toolbartype) {
 case 1 : gtk_toolbar_set_style(tb, GTK_TOOLBAR_BOTH); break;
 case 2 : gtk_toolbar_set_style(tb, GTK_TOOLBAR_TEXT); break;
 default : gtk_toolbar_set_style(tb, GTK_TOOLBAR_ICONS); break;
 }
 if (gnomamerc->showinfo) {
    b = GTK_TOGGLE_BUTTON(lookup_widget(widget, "infob"));
    gtk_toggle_button_set_active(b, TRUE);
 }
 if (gnomamerc->showpreview) {
    b = GTK_TOGGLE_BUTTON(lookup_widget(widget, "previewb"));
    gtk_toggle_button_set_active(b, TRUE);
 }
}


void
on_updateb_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
 gnomame_query_xmame(mainw);
}


void
on_filterb_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkEntry *description, *name, *year, *manufacturer, *clone;

 description = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "description"));
 name = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "name"));
 year = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "year"));
 manufacturer = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "manufacturer"));
 clone = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "clone"));

 get_entry(description, &filter.description);
 get_entry(name, &filter.name);
 get_entry(year, &filter.year);
 get_entry(manufacturer, &filter.manufacturer);
 get_entry(clone, &filter.cloneof);
 gnomame_filter();
}



void
on_closeb_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
 g_free(filter.description);
 g_free(filter.cloneof);
 g_free(filter.name);
 g_free(filter.year);
 g_free(filter.manufacturer);
 filter.description = filter.cloneof = filter.name = NULL;
 filter.manufacturer = filter.year = NULL;
 gnomame_filter();
 gtk_widget_hide(GTK_WIDGET(user_data));
}


void
on_helpb_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
 gnomame_show_help(user_data);
}




void
on_notepadb_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_widget_show(notepad);
}




void notepad_updatestatus(void)

{
 GtkCList *notelist = GTK_CLIST(lookup_widget(notepad, "notelist"));
 GnomeAppBar *bar = GNOME_APPBAR(lookup_widget(notepad, "notestatus"));
 gchar s[256];

 if (notelist->rows)
	 sprintf(s, "%d marked games", notelist->rows);
 else
	 s[0] = 0;

 gnome_appbar_set_status(bar, s);
}



void
on_noteremoveb_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkCList *l = GTK_CLIST(user_data);
 gint i;
 gamerecord *g;
 
 gtk_clist_freeze(l);
 while (l->selection != NULL) {
	 i = GPOINTER_TO_INT(l->selection->data);
	 g = gtk_clist_get_row_data(l, i);
    g->marked = FALSE;
	 gnomame_refresh_game(g);
	 gtk_clist_remove(l, i);
 }
 gtk_clist_thaw(l);
 notepad_updatestatus();
}


void
on_noteclearb_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 gint i;
 gamerecord *g;

 for (i = 0; i < GTK_CLIST(user_data)->rows; i++) {
	 g = gtk_clist_get_row_data(GTK_CLIST(user_data), i);
    g->marked = FALSE;
	 gnomame_refresh_game(g);
 }
 gtk_clist_clear(GTK_CLIST(user_data));
 notepad_updatestatus();
}


void
on_notecloseb_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_widget_hide(notepad);
}


void
on_autoframeskip_toggled               (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "autoframeskip", TRUE);
}



void
on_keepaspect_toggled                  (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "keepaspect", TRUE);
}



void
on_xil_toggled                         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "xil", TRUE);
}




void
on_mapkeyadd_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkEntry *lentry, *rentry;
 GtkCList *mapkeys;
 gchar *row[2];

 lentry = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "w_mapkey_lentry"));
 rentry = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "w_mapkey_rentry"));
 mapkeys = GTK_CLIST(user_data);
 row[0] = gtk_entry_get_text(lentry);
 row[1] = gtk_entry_get_text(rentry);
 if (row[0][0] != '\0' && row[1][0] != '\0') {
    gtk_clist_append(mapkeys, row);
    gtk_entry_set_text(lentry, "");
    gtk_entry_set_text(rentry, "");
 }
}


void
on_mapkeydelete_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkCList *l;

 l = GTK_CLIST(user_data);
 while (l->selection)
	 gtk_clist_remove(l, GPOINTER_TO_INT(l->selection->data));
}


void
on_disablemaddb_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkEntry *entry;
 GtkCList *l;
 gchar *row[1];

 entry = GTK_ENTRY(lookup_widget(GTK_WIDGET(button), "w_disablemode_entry"));
 l = GTK_CLIST(user_data);
 row[0] = gtk_entry_get_text(entry);
 if (row[0][0] != '\0') {
    gtk_clist_append(l, row);
    gtk_entry_set_text(entry, "");
 }
}


void
on_disablemdelb_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkCList *l;

 l = GTK_CLIST(user_data);
 while (l->selection)
	 gtk_clist_remove(l, GPOINTER_TO_INT(l->selection->data));
}



static void filesel_saveas_click(GtkWidget *w, GtkCList *clist)

{
 GtkFileSelection *fs = GTK_FILE_SELECTION(gtk_widget_get_toplevel(w));
 gchar s[256], *s1;
 gamerecord *g;
 FILE *f;
 gint i;

 if (fs->ok_button == w) {
	 if ((f = fopen(s1 = gtk_file_selection_get_filename(fs), "w"))) {
	    for (i = 0; i < clist->rows; i++) {
		    g = gtk_clist_get_row_data(clist, i);
			 fprintf(f, "%s;%s\n", g->name, g->description);
       }
		 fclose(f);
    } else {
		 sprintf(s, "Could not open file %s for write", s1);
	    gnomame_error(s, 0);
	 }
 }
 gtk_widget_destroy(GTK_WIDGET(fs));
}


void
on_notesaveas_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
 GtkFileSelection *filesel;

 filesel = GTK_FILE_SELECTION(gtk_file_selection_new("Save marked as..."));
 gtk_signal_connect(GTK_OBJECT(filesel),
		              "destroy", (GtkSignalFunc) gtk_widget_destroy,
						  filesel);
 gtk_signal_connect(GTK_OBJECT(filesel->ok_button),
			           "clicked", (GtkSignalFunc) filesel_saveas_click,
						  user_data);
 gtk_signal_connect(GTK_OBJECT(filesel->cancel_button),
	                 "clicked", (GtkSignalFunc) filesel_saveas_click,
						  user_data);
 gtk_window_set_modal(GTK_WINDOW(filesel), TRUE);
 gtk_widget_show(GTK_WIDGET(filesel));

}


void
on_add_to_notepad1_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_mark_selected(GTK_CLIST(lookup_widget(notepad, "notelist")));
 gnomame_unselect_all();
}


void
on_remove_from_notepad1_activate       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_unmark_selected(GTK_CLIST(lookup_widget(notepad, "notelist")));
}




gboolean
on_notelist_button_press_event         (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
 GtkCList *l = GTK_CLIST(widget);
 gint row, col, ret;
 gamerecord *g;

 if (widget && event) {
	 if (event->button == 1 && event->type == GDK_2BUTTON_PRESS) {
		 ret = gtk_clist_get_selection_info(l, event->x, event->y, &row, &col);
		 if (ret) {
		    g = gtk_clist_get_row_data(l, row);
			 gnomame_goto_game(g->name);
	    }
	    return (TRUE);
    }
 }

 return (FALSE);
}


void
on_delete_preferences1_activate        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_delrc();
}


void
on_preferences1_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 open_xmamepropw();
}


void
on_notehelpb_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
 gnomame_show_help(user_data);
}


void
on_clear_selection1_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_unselect_all();
}


void
on_select_all1_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_select_all();
}

gboolean
on_searchw_delete_event                (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 on_closeb_clicked(NULL, widget);

 return TRUE;
}


void
on_listw_click_column                  (GtkCList        *clist,
                                        gint             column,
                                        gpointer         user_data)
{
 if (column == 0)
    gnomame_cycle_filter(clist);
 else {
    if (column == clist->sort_column) {
       if (clist->sort_type == GTK_SORT_ASCENDING)
          clist->sort_type = GTK_SORT_DESCENDING;
       else
          clist->sort_type = GTK_SORT_ASCENDING;
    } else
       gtk_clist_set_sort_column(clist, column);
    gtk_clist_sort(clist);
 }
}


void
on_listw_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 gnomame_change_selection(clist);
}


void
on_listw_unselect_row                  (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
 gnomame_change_selection(clist);
}




gboolean
on_listw_key_press_event               (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{
 GtkToggleButton *b;

 if (widget && event) {
    switch (gdk_keyval_to_upper(event->keyval)) {
	 case GDK_Return :
       if (gnomame_selectedgame) gnomame_play_game(gnomame_selectedgame);
		 break;
	 case GDK_U :
       gnomame_unmark_selected(GTK_CLIST(lookup_widget(notepad, "notelist")));
		 break;
	 case GDK_M :
       gnomame_mark_selected(GTK_CLIST(lookup_widget(notepad, "notelist")));
       gnomame_unselect_all();
		 break;
	 case GDK_P :
		 open_xmamepropw();
		 break;
	 case GDK_H :
		 gnomame_show_help(NULL);
		 break;
	 case GDK_S :
       b = GTK_TOGGLE_BUTTON(lookup_widget(widget, "previewb"));
       gtk_toggle_button_set_active(b, !b->active);
		 break;
	 case GDK_I :
       b = GTK_TOGGLE_BUTTON(lookup_widget(widget, "infob"));
       gtk_toggle_button_set_active(b, !b->active);
		 break;
	 case GDK_space :
		 gtk_signal_emit_by_name(GTK_OBJECT(widget), "scroll_vertical",
				                   GTK_SCROLL_STEP_FORWARD, 0);
		 break;
	 case GDK_BackSpace :
		 gtk_signal_emit_by_name(GTK_OBJECT(widget), "scroll_vertical",
				                   GTK_SCROLL_STEP_BACKWARD, 0);
		 break;
    default :
		 return (FALSE);
		 break;
    }
	 return (TRUE);
 }

 return (FALSE);
}




gboolean
on_listw_button_press_event            (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
 extern GtkWidget *gamesmenu;
 GtkCList *l = GTK_CLIST(widget), *notelist;
 gint row, col, ret = 0;
 gamerecord *game = NULL;

 if (widget && event) {
	 if (event->button != 1) {
		 ret = gtk_clist_get_selection_info(l, event->x, event->y, &row, &col);
		 if (l->selection == NULL && ret) gtk_clist_select_row(l, row, col);
		 if (l->selection) {
	       if (!gtk_clist_row_is_selected(l, row)) {
				 gtk_clist_unselect_all(l);
	          gtk_clist_select_row(l, row, col);
          }
          gtk_clist_set_focus_row(l, row);
			 game = (gamerecord *) gtk_clist_get_row_data(l, row);
       }
	 }
	 switch (event->button) {
	 case 1 :
		 if (event->type == GDK_2BUTTON_PRESS && gnomame_selectedgame)
			 gnomame_play_game(gnomame_selectedgame);
	    else
			 return (FALSE);
		 break;
	 case 3 :
		 if (event->type == GDK_BUTTON_PRESS)
			 gtk_menu_popup(GTK_MENU(gamesmenu), NULL, NULL, NULL, NULL,
					          event->button, event->time);
       break;
	 case 2 :
		 notelist = GTK_CLIST(lookup_widget(notepad, "notelist"));
       if (game && game->marked)
			 gnomame_unmark_selected(notelist);
		 else
			 gnomame_mark_selected(notelist);
		 break;
    default :
		 return (FALSE);
		 break;
	 }
	 return (TRUE);
 }
 return (FALSE);
}




static gboolean check_enter(GtkWidget *w, GdkEventKey *e, gpointer data)

{
 if (w && e) {
	 if (e->keyval == GDK_Return) {
	    on_filterb_clicked(GTK_BUTTON(lookup_widget(w, "filterb")), NULL);
	    return (TRUE);
	 } else if (e->keyval == GDK_Escape) {
	    on_closeb_clicked(GTK_BUTTON(lookup_widget(w, "closeb")),
				             lookup_widget(w, "searchw"));
	    return (TRUE);
    }
 }

 return (FALSE);
}



void on_searchw_realize(GtkWidget *widget, gpointer user_data)

{
 GtkWidget *w[5];
 gint i = 0;

 w[i++] = lookup_widget(widget, "description");
 w[i++] = lookup_widget(widget, "name");
 w[i++] = lookup_widget(widget, "year");
 w[i++] = lookup_widget(widget, "manufacturer");
 w[i++] = lookup_widget(widget, "clone");

 while (i-- > 0)
	 gtk_signal_connect(GTK_OBJECT(w[i]), "key_press_event",
			              GTK_SIGNAL_FUNC(check_enter), NULL);
}

void
on_button13_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
 gnomame_show_help(user_data);
}

void
on_w_disablemode_clear_clicked         (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_clist_clear(GTK_CLIST(user_data));
}


void
on_w_mapkey_clear_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
 gtk_clist_clear(GTK_CLIST(user_data));
}


void
on_go_to_clone_of_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 if (gnomame_selectedgame) gnomame_goto_game(gnomame_selectedgame->cloneof);
}


void
on_general_options1_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gnomame_unselect_all();
 gtk_widget_show(create_xmamepropw());
}


void
on_game_specific_options1_activate     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
 gtk_widget_show(create_xmamepropw());
}


void
on_w_autodouble_toggled                (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "autodouble", TRUE);
}


void
on_w_samples_toggled                   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "samples", TRUE);
}


void
on_w_mouse_toggled                     (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "mouse", TRUE);
}



void
on_w_cabview_toggled                   (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(togglebutton), "cabview", TRUE);
}



void
on_combo_entry9_changed                (GtkEditable     *editable,
                                        gpointer         user_data)
{
 set_sensitivity(GTK_WIDGET(editable), "joytype", TRUE);
}
