#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gnome.h>
#include <ctype.h>

#include "support.h"
#include "preferences.h"
#include "core.h"
#include "callbacks.h"
#include "mdsgtkhelp.h"



void
on_listw_click_column                  (GtkCList        *clist,
                                        gint             column,
                                        gpointer         user_data);

void
on_listw_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data);

void
on_listw_unselect_row                  (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data);

gboolean
on_listw_key_press_event               (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data);

gboolean
on_listw_button_press_event            (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data);

static gboolean 
is_directory							(const gchar *directory, 
										 const gchar *file);

#define skip_until(ptr, c) { \
	while (*ptr != '\0' && *ptr != c) ptr++; \
   while (*ptr != '\0' && *ptr == c) ptr++; }




typedef struct {
	GdkPixmap *pix;
	GdkBitmap *mask;
} gnomame_pix;




typedef struct {
	gchar *code;
	gchar *description;
} xmame_plugin;




typedef struct {
	gchar *name;
	gchar *code;
	gint width;
} gameslist_col;




#define MAXCOLNUM 8


static gameslist_col listw_col[MAXCOLNUM];
static gint listw_colnum;
xmameprefs *xmamerc;
GSList *gnomame_mixerplugins = NULL, *gnomame_dspplugins = NULL;
static GdkFont *infofont = NULL, *errorfont = NULL;
static gnomame_pix noavailpix, availpix, qmarkpix, redqmarkpix, rcfile, marker;
GtkCList *listw = NULL;
static int ngames = 0;
static gamerecord **games = NULL;
static gfloat progress = 1;
static gfloat progressincr = -0.01;
gamerecord *gnomame_selectedgame = NULL;
extern GtkWidget *xmamepropw;
extern GtkWidget *notepad;
extern GtkWidget *gamesmenu;
static gchar *xmameversion;



extern void notepad_updatestatus(void);



static gboolean exist_file(gchar *s)

{
 FILE *f;

 if ((f = fopen(s, "r"))) {
	 fclose(f);
	 return (TRUE);
 } else
	 return (FALSE);
}



static gboolean is_playable(gamerecord *g)

{
 if (g == NULL) return (FALSE);
 else return (g->state <= GNOMAME_WARNING);
}




static FILE *gnomame_pipe(gchar *cmd, gchar *name)

{
 gchar *s, s1[256];
 FILE *f;

 if (name == NULL) name = "";
 s = g_strconcat(gnomamerc->xmamepath,
		           " ", cmd, " ", name, " 2>/dev/null", NULL);
 if ((f = popen(s, "r")) == NULL) {
    sprintf(s1, "Couldn't execute xmame command : %s, popen failed", cmd);
	 gnomame_error(s1, GNOMAME_WARNING);
 }
 g_free(s);

 return (f);
}




static void load_pixmap(gnomame_pix *p, gchar *name, GtkWidget *w)

{
 GtkStyle *style;
 gchar *s, *s1;
 
 if (p->pix == NULL && name != NULL) {
    style = gtk_widget_get_style(w);
    s = gnome_pixmap_file(s1 = g_strconcat("/gnomame/", name, NULL));
	 g_free(s1);
    p->pix = gdk_pixmap_create_from_xpm(w->window, &p->mask,
   		                               &style->bg[GTK_STATE_NORMAL], s);
    g_free(s);
 }
}




static void load_pixmaps(GtkWidget *w)

{
 load_pixmap(&availpix, "greenv.xpm", w);
 load_pixmap(&noavailpix, "redx.xpm", w);
 load_pixmap(&qmarkpix, "qmark.xpm", w);
 load_pixmap(&redqmarkpix, "redqmark.xpm", w);
 load_pixmap(&rcfile, "rcfile.xpm", w);
 load_pixmap(&marker, "marker.xpm", w);
}




static void gnomame_set_status(gchar *s)

{
 gnome_appbar_set_status(GNOME_APPBAR(lookup_widget(GTK_WIDGET(listw),
				                                        "mainstatus")), s);
 while (gtk_events_pending()) gtk_main_iteration() ;
}




static void gnomame_set_progress(gfloat percent)
	
{
 if (percent > 1.0) percent = 1.0;
 gnome_appbar_set_progress(GNOME_APPBAR(lookup_widget(GTK_WIDGET(listw),
				                                          "mainstatus")),
                           percent);
 while (gtk_events_pending()) gtk_main_iteration() ;
}




static void gnomame_update_progress(void)

{
 if (progress < 0 || progress > 1) {
	 progressincr = -progressincr;
	 progress += progressincr;
 }
 gnomame_set_progress(progress);
 progress += progressincr;
}



void gnomame_update_status(void)

{
 GnomeAppBar *bar = GNOME_APPBAR(lookup_widget(GTK_WIDGET(listw),
			                                      "mainstatus"));
 gchar s[256];

 sprintf(s, "%s - Total : %d    Displayed %d",
         (xmameversion != NULL) ? xmameversion : "No xmame",
         ngames, listw->rows);
 gnome_appbar_set_status(bar, s);
 if (ngames)
    gnome_appbar_set_progress(bar, (gfloat) listw->rows / (gfloat) ngames);
 else
	 gnome_appbar_set_progress(bar, 0);
 while (gtk_events_pending()) gtk_main_iteration() ;
}




void gnomame_error(char *msg, int type)

{
 GtkWidget *w;
 gchar *s;

 
 if (type == GNOMAME_FATAL) {
	 s = g_strconcat("Fatal error : ", msg, NULL);
    w = gnome_message_box_new(s, GNOME_MESSAGE_BOX_ERROR,
			                     GNOME_STOCK_BUTTON_CLOSE, NULL);
 } else {
	 s = g_strconcat("Warning : ", msg, NULL);
    w = gnome_message_box_new(s, GNOME_MESSAGE_BOX_INFO,
			                     GNOME_STOCK_BUTTON_CLOSE, NULL);
 }
 g_free(s);
 gnome_dialog_run_and_close(GNOME_DIALOG(w));
 if (type == GNOMAME_FATAL) exit(1);
}




static gboolean wild_compare(gchar *s, gchar *key)

{
 if (key == NULL || *key == 0) return (TRUE);
 if (s == NULL || *s == 0) return (FALSE);
 while (*s != 0 && *key != 0) {
	 if (*key == '?' || tolower(*s) == tolower(*key)) {
		 s++; key++; 
	 } else return (FALSE);
 }
 return (*key == 0);
}



static gboolean match_strings(gchar *s, gchar *key)

{
 gchar s1[256], *sx;
 int i, end;
 gboolean flag;

 if (s == NULL) s = "";
 if (key == NULL || *key == 0 || strcmp(key, "*") == 0) return (TRUE);
 else {
	 flag = (*key == '*');
	 while (*key == '*') key++;
	 for (sx = s1; *key != '*' && *key != 0; sx++, key++) *sx = *key;
	 *sx = 0;
	 end = (flag) ? strlen(s) - strlen(s1) : 0;
	 for (i = 0; i <= end; i++)
		 if (wild_compare(s + i, s1)) {
			 sx = s + i + strlen(s1);
          if (*key != 0) 
			    return match_strings(sx, key);
			 else
				 return (*sx == 0);
       }
 }
 return (FALSE);
}


static void gamerecord_free(gamerecord *g)

{
 g_slist_foreach(g->errors, (GFunc) g_free, 0);
 g_slist_free(g->errors);
 if (g->name) g_free(g->name);
 if (g->description) g_free(g->description);
 if (g->manufacturer) g_free(g->manufacturer);
 if (g->year) g_free(g->year);
 if (g->cloneof) g_free(g->cloneof);
 g_free(g);
}



/* Return a list with the names of playable games
 *
 * Warning : There is no need to free the names, free only the list itself
 * because the names are referenced into the main games list */

GList *gnomame_give_availables(void)

{
 gamerecord *g;
 GList *l = NULL;
 int i;

 if (games) {
    for (i = 0; i < ngames; i++) {
       g = games[i];
	    if (is_playable(g)) l = g_list_append(l, g->name);
    }
 }
 if (l == NULL) l = g_list_append(l, "pacman");

 return (l);
}





static int g_strcmp(gchar *s1, gchar *s2)

{
      if (s1 == NULL && s2 == NULL) return (0);
 else if (s1 == NULL && s2 != NULL) return (-1);
 else if (s1 != NULL && s2 == NULL) return (1);
 else return (strcmp(s1, s2));
}





static GSList *check_rom(gchar *name, int *state)

{
 GSList *l = NULL;
 FILE *f;
 gchar s[256], *s1, s2[256], *s3;

 *state = 0;
 s[255] = 0;
 f = gnomame_pipe("-verifyroms", name);
 while (!feof(f)) {
	 if ((s1 = fgets(s, 255, f)) != s) break;
	 if (strncmp(s1, "romset", 6) == 0) {
		 skip_until(s1, ' ');
		 skip_until(s1, ' ');
		 s3 = s2;
		 while (*s1 != '\0' && *s1 != '\n') *s3++ = *s1++;
		 *s3 = '\0';
		 if (*state == 0) {
			 if (strcmp(s2, "correct") == 0) *state = GNOMAME_OK;
			 else if (strcmp(s2, "incorrect") == 0) *state = GNOMAME_WARNING;
			 else if (strcmp(s2, "not found") == 0) {
				 *state = GNOMAME_FATAL;
				 g_slist_foreach(l, (GFunc) g_free, NULL);
				 g_slist_free(l);
				 l = g_slist_append(NULL, g_strdup("Zip file corrupt\n"));
          }
       }
		 break;
    } else {
		 skip_until(s1, ':');
		 skip_until(s1, ' ');
		 if (strstr(s1, "NOT FOUND")) *state |= GNOMAME_FATAL;
    }
	 l = g_slist_append(l, g_strdup(s1));
 }
 pclose(f);

 return (l);
}





static int cmpfunc(gamerecord *g1, gamerecord *g2)

{
 return (strcmp(g1->name, g2->name));
}




/* check_availables() : set the status flag for the availables games
 * founded into ROM directories */




static GSList *check_availables(GSList *gameslist, GSList *romdirs)

{
 gamerecord *g;
 GSList *l, *l1, *l2, *start;
 struct dirent *d;
 DIR *D;
 gchar *s, *home;
 int i, cnt = 0;

 /* Find in all the subdirectories of all the rompaths for roms and store
  * the rom name founded in a sorted list without duplicates              */

 l = NULL;
 while (romdirs) {
	 s = g_strconcat("Scanning dir. for ROMs : ", romdirs->data, NULL);
    gnomame_set_status(s);
	 g_free(s);
	 if ((D = opendir(romdirs->data))) {
	    while ((d = readdir(D))) {
			 gnomame_update_progress();
			 if (d->d_name[0] == '.') continue;
	   	 i = strlen(d->d_name);
			 if (i > 12) continue;
			 if (!is_directory(romdirs->data, d->d_name)
				  && d->d_type != DT_REG
				  && d->d_type != DT_UNKNOWN) continue;
			 if (d->d_type == DT_DIR && i > 8) continue;
			 if ((d->d_type == DT_REG || d->d_type == DT_UNKNOWN)
				  && i < 5) continue;
		    g_strdown(d->d_name);
			 if (!is_directory(romdirs->data, d->d_name) 
              && strcmp(d->d_name + i - 4, ".zip") != 0) continue;
			 if (is_directory(romdirs->data, d->d_name))
				 s = g_strdup(d->d_name);
			 else
				 s = g_strndup(d->d_name, i - 4);
			 if (g_slist_find_custom(l, s, (GCompareFunc) strcmp))
				 g_free(s);
			 else
				 l = g_slist_insert_sorted(l, s, (GCompareFunc) strcmp);
       }
	    closedir(D);
    }
	 romdirs = romdirs->next;
 }

 /* In this second part we iterate for every rom in the list finding the
  * corresponding gamerecord in the main games list (also sorted) */
 home = g_strconcat(g_get_home_dir(), "/.xmame/rc/", NULL);

 gameslist = g_slist_sort(gameslist, (GCompareFunc) cmpfunc);

 start = gameslist;
 for (l1 = l; l1; l1 = l1->next) {
	 for (l2 = start; l2; l2 = l2->next) {
		 g = (gamerecord *) l2->data;
	    if ((cnt++ % 10) == 0) gnomame_update_progress();
		 if (strcmp(g->name, l1->data) == 0) {
		    gnomame_set_status(s = g_strconcat("Testing ROM : ",
						                           g->name, " ...", NULL));
		    g_free(s);
          g->haverc = exist_file(s = g_strconcat(home, g->name, "rc", NULL));
          g_free(s);
		    g_slist_foreach(g->errors, (GFunc) g_free, 0);
		    g_slist_free(g->errors);
		    g->errors = check_rom(l1->data, &g->state);
			 start = l2->next;
			 break;
       }
    }
 }
 g_free(home);
 g_slist_free(l);
 return (gameslist);
}





static gboolean filter_match(gamerecord *g)

{
 if ((g->state & filter.state) == 0) return (FALSE);
 if (!match_strings(g->name, filter.name)) return (FALSE);
 if (!match_strings(g->description, filter.description)) return (FALSE);
 if (!match_strings(g->year, filter.year)) return (FALSE);
 if (!match_strings(g->cloneof, filter.cloneof)) return (FALSE);
 if (!match_strings(g->manufacturer, filter.manufacturer)) return (FALSE);

 return (TRUE);
}




static void gameslist_free(void)

{
 int i;

 if (games != NULL) {
	 for (i = 0; i < ngames; i++) gamerecord_free(games[i]);
	 g_free(games);
	 games = NULL;
 }
}





static gchar *gnomame_xmame_version(void)

{
 gchar *s, s1[256];
 FILE *f;

 f = gnomame_pipe("-version", "");
 s = fgets(s1, 255, f);
 pclose(f);
 if (s == s1) {
	 while (*s != '\0' && *s != '\n') s++;
	 if (*s == '\n') *s = '\0';
	 return (g_strdup(s1));
 }
 return (NULL);
}




gboolean gnomame_check_option(gchar *option)

{
 gchar *s, riga[256], *fatal;
 FILE *f;
 gboolean flag = FALSE;
 int i;

 fatal = g_strconcat("error: unknown option -", option, NULL);
 s = g_strconcat(gnomamerc->xmamepath, " -version -", option, " 2>&1", NULL);
 if ((f = popen(s, "r")) == NULL)
    gnomame_error("Couldn't execute xmame to check options", GNOMAME_FATAL);
 g_free(s);
 riga[255] = 0;
 while (!feof(f) && !flag) {
    s = fgets(riga, 255, f);
	 if (s != riga) break;
	 if (riga[i = strlen(riga)] == '\n') riga[i] = '\0';
	 g_strchomp(riga);
	 flag = (strcmp(riga, fatal) == 0);
 }
 pclose(f);

 return (!flag);
}





static GSList *gnomame_givegames(int *total)

{
 FILE *f;
 char *s, riga[256], name[9], descr[256];
 GSList *l = NULL;
 gamerecord *gr;
 int i;

 *total = 0;
 f = gnomame_pipe("-listfull", "");
 riga[255] = 0;
 descr[255] = 0;
 name[8] = 0;
 s = fgets(riga, 255, f);
 if (s != riga) return (NULL);
 while (!feof(f)) {
    s = fgets(riga, 255, f);
    if (s != riga) break;
	 if (*s == '\0' || *s == '\n') break;
	 if (*s != '-') {
		 strncpy(name, s, 8);
		 strncpy(descr, s + 10, 255);
       i = strlen(descr) - 1;
		 if (descr[i] == '\n') descr[i] = '\0';
		 if (*name) {
          l = g_slist_append(l, gr = g_malloc0(sizeof(gamerecord)));
          if (gr == NULL)
             gnomame_error("Failed to allocate memory for gamerecord",
		                     GNOMAME_FATAL);
          g_strchomp(name);
          gr->name = g_strdup(name);
          g_strchomp(descr);
          if (descr[0] == '"')
             gr->description = g_strndup(descr + 1, strlen(descr + 1) - 1);
          else
             gr->description = g_strdup(descr);
          gr->state = GNOMAME_NOROM;
	       (*total)++;
		 }
    }
 }
 pclose(f);

 return (l);
}




static int search_column(char *s1, char *s2)

{
 int i = -1;

 s2 = strstr(s1, s2);
 if (s2) for (i = 0; s1 != s2; i++) s1++;
 return (i);
}





/* read_gameslist() : call xmame to retrieve the most useful informations from
 * games */


static GSList *gnomame_readgames(gint *ngames)

{
 gamerecord *gr = NULL;
 FILE *f;
 char *s, *ss;
 char s1[256], riga[256];
 int  cl[4], ng;
 GSList *l, *l1;

 gnomame_set_status("Reading games info");
 l = gnomame_givegames(&ng);
 f = gnomame_pipe("-listgames", "");
 riga[255] = 0;
 while (!feof(f)) {
    s = fgets(riga, 255, f);
    if (s != riga) break;
	 if (*s && *s != '-') break; /* Skip empty and '------' lines */
 }
 /* Here we have found the title so we can search for columns,
    standard search should find : name/year/manufacturer/description */
 g_strdown(riga);
 cl[0] = search_column(riga, "name");
 cl[1] = search_column(riga, "year");
 cl[2] = search_column(riga, "manufacturer");
 cl[3] = search_column(riga, "description");
 if (cl[3] == -1) {
 /* Fix for xmame >= 0.37b10.1 : year/manufacturer/name */
    cl[3] = cl[0];
    cl[0] = -1;
 }
 l1 = l;
 while (!feof(f) && l1) {
    s = fgets(riga, 255, f);
    if (s != riga || *s == '\n') break;
	 if (*s != '-') {
		 gr = (gamerecord *) l1->data;
		 if (cl[1] >= 0) gr->year = g_strndup(riga + cl[1], 4);
		 if (cl[2] >= 0 && cl[3] >= 0) {
			 gr->manufacturer = g_strndup(riga + cl[2], cl[3] - cl[2]);
		    g_strchomp(gr->manufacturer);
       }
		 l1 = l1->next;
    }
 }
 pclose(f);
 f = gnomame_pipe("-listclones", "");
 s1[255] = 0;
 s = fgets(riga, 255, f);
 l1 = l;
 while (!feof(f)) {
    s = fgets(riga, 255, f);
    if (s != riga) break;
	 while (*s == ' ' || *s == '\t') s++;
	 for (ss = s1; isalnum(*s); s++, ss++) *ss = *s;
	 *ss = 0;
	 while (l) {      /* -listgames and -listclones have the same order so we can
							 * search in this way */
		 gr = (gamerecord *) l->data;
		 l = l->next;
		 if (strcmp(gr->name, s1) == 0) {
			 while (*s == ' ' || *s == '\t') s++;
			 for (ss = s1; isalnum(*s); s++, ss++) *ss = *s;
			 *ss = 0;
			 gr->cloneof = g_strdup(s1);
			 break;
       }
    }
 }
 pclose(f);
 *ngames = ng;

 return l1;
}



void gnomame_load_games(void)

{
 GSList *l, *l1;
 int i, ng;
 xmamepref *rompath;

 ngames = 0;
 l = gnomame_readgames(&ng);
 rompath = xmameprefs_find(xmamerc, "rompath");
 l = check_availables(l, rompath->val.l);
 games = (gamerecord **) g_new(gamerecord *, ng);
 for (l1 = l, i = 0; l1; l1 = l1->next, i++) games[i] = l1->data;
 ngames = ng;
 gnomame_filter();
 g_slist_free(l);
}




static void gnomame_game_to_row(GtkCList *l, gint r, gamerecord *g)

{
 gchar *row[MAXCOLNUM];
 gint c = 0;

 g->row = r;
 row[c++] = NULL;
 row[c++] = NULL;
 row[c++] = NULL;
 row[c++] = g->description;
 if (gnomamerc->showcol[0]) row[c++] = g->name;
 if (gnomamerc->showcol[1]) row[c++] = g->year;
 if (gnomamerc->showcol[2]) row[c++] = g->manufacturer;
 if (gnomamerc->showcol[3]) row[c++] = g->cloneof;
 for (c = 0; c < listw_colnum; c++) gtk_clist_set_text(l, r, c, row[c]);
 switch (g->state) {
 case GNOMAME_OK :
    gtk_clist_set_pixmap(l, r, 0, availpix.pix, availpix.mask); break;
 case GNOMAME_FATAL :
    gtk_clist_set_pixmap(l, r, 0, redqmarkpix.pix, redqmarkpix.mask); break;
 case GNOMAME_WARNING :
    gtk_clist_set_pixmap(l, r, 0, qmarkpix.pix, qmarkpix.mask); break;
 default :
    gtk_clist_set_pixmap(l, r, 0, noavailpix.pix, noavailpix.mask); break;
 }
 if (g->haverc) gtk_clist_set_pixmap(l, r, 1, rcfile.pix, rcfile.mask);
 if (g->marked) gtk_clist_set_pixmap(l, r, 2, marker.pix, marker.mask);
 gtk_clist_set_row_data(l, r, (gpointer) g);
}




static void insert_game(GtkCList *l, gint r, gamerecord *g)

{
 gchar *row[MAXCOLNUM];
 gint i;

 for (i = 0; i < listw_colnum; i++) row[i] = 0;
 gtk_clist_insert(l, r, row);
 gnomame_game_to_row(l, r, g);
}




/* gnomame_filter : take from games and put those match the filter into listw
 * */

void gnomame_filter(void)

{
 gint i, row, r;

 row = gtk_clist_row_get_first(listw);
 gnomame_set_status("Filtering games");
 gtk_clist_freeze(listw);
 gtk_clist_clear(listw);
 for (i = 0, r = 0; i < ngames; i++)
	 if (filter_match(games[i])) insert_game(listw, r++, games[i]);
 gtk_clist_thaw(listw);
 gnomame_update_status();
 gnomame_change_selection(listw);
 if (row >= 0) gtk_clist_moveto(listw, row, 0, 0, 0);
}




static void gnomame_refresh_row(gint r)

{
 gnomame_game_to_row(listw, r, gtk_clist_get_row_data(listw, r));
}




void gnomame_refresh_game(gamerecord *g)

{
 gnomame_refresh_row(g->row);
}



void gnomame_refresh_selected(void)

{
 GList *l1;

 gtk_clist_freeze(listw);
 for (l1 = listw->selection; l1; l1 = l1->next)
	 gnomame_refresh_row(GPOINTER_TO_INT(l1->data));
 gtk_clist_unselect_all(listw);
 gtk_clist_thaw(listw);
}




void gnomame_refresh_all(void)

{
 gint i;

 gtk_clist_freeze(listw);
 for (i = 0; i < listw->rows; i++) gnomame_refresh_row(i);
 gtk_clist_thaw(listw);
}



void gnomame_delrc(void)

{
 GList *l1;

 gtk_clist_freeze(listw);
 for (l1 = listw->selection; l1; l1 = l1->next) {
    xmameprefs_delrc(gtk_clist_get_row_data(listw,
				                                GPOINTER_TO_INT(l1->data)));
	 gnomame_refresh_row(GPOINTER_TO_INT(l1->data));
 }
 gtk_clist_unselect_all(listw);
 gtk_clist_thaw(listw);
}




static void set_first_column(GtkCList *list, int state)

{
 GtkWidget *w;

 if ((w = gtk_clist_get_column_widget(list, 0))) gtk_widget_destroy(w);
 if (state & GNOMAME_NOROM)
	 w = gtk_pixmap_new(noavailpix.pix, noavailpix.mask);
 else if (state & GNOMAME_FATAL)
	 w = gtk_pixmap_new(redqmarkpix.pix, redqmarkpix.mask);
 else if (state & GNOMAME_WARNING)
	 w = gtk_pixmap_new(qmarkpix.pix, qmarkpix.mask);
 else if (state & GNOMAME_OK)
	 w = gtk_pixmap_new(availpix.pix, availpix.mask);
 gtk_clist_set_column_widget(list, 0, w);
}




void gnomame_cycle_filter(GtkCList *l)

{
 if (filter.state & GNOMAME_NOROM) filter.state = GNOMAME_OK;
 else if (filter.state & GNOMAME_FATAL) filter.state |= GNOMAME_NOROM;
 else if (filter.state & GNOMAME_WARNING) filter.state |= GNOMAME_FATAL;
 else filter.state |= GNOMAME_WARNING;
 set_first_column(l, filter.state);
 gnomame_filter();
}




static void real_play(gint reply, gamerecord *game)

{
 pid_t pid;
 FILE *f;
 gchar *s;

 if (reply == 0) {
    pid = fork();
    if (pid == 0) {
/*       s = g_strconcat(gnomamerc->xmamepath, " ",
                     game->name, ">/dev/null 2>/dev/null", NULL); */
       s = g_strconcat(gnomamerc->xmamepath, " ",
                     game->name, " >/dev/null 2>&1", NULL);
       f = popen(s, "w");
       fprintf(f, "\n"); /* Sometimes on errors xmame ask a confirmation */
       pclose(f);
       g_free(s);
       _exit(0);
    } else if (pid == -1)
       gnomame_error("fork() failed", GNOMAME_WARNING);
 }
}




void gnomame_play_game(gamerecord *game)

{
 if (is_playable(game)) {
	 if (gnomamerc->confirmplay) {
       gchar *s;

		 s = g_strconcat("Play ", game->description, " ?", NULL);
	    gnome_ok_cancel_dialog_modal(s, (GnomeReplyCallback) real_play,
				                        (gpointer) game);
		 g_free(s);
    } else
		 real_play(0, game);
 }
}




static GSList *gnomame_readgameinfo(gamerecord *g)

{
 GSList *l = NULL;
 char *s, s1[256], s2[256];
 char rheader[256], rdelim[256], rinfo[256];
 char *rh, *rd, *ri;
 int i, i1;
 FILE *f;

 l = g_slist_append(l, g_strconcat("Game: ", g->description, "\n", NULL));
 l = g_slist_append(l, g_strconcat("Year: ", g->year, "\n", NULL));
 l = g_slist_append(l, g_strconcat("Manufacturer: ", g->manufacturer, "\n", NULL));
 f = gnomame_pipe("-listdetails", g->name);
 rheader[255] = 0;
 s = fgets(rheader, 255, f);
 if (!feof(f) && s == rheader) { /* Read header */
    rdelim[255] = 0;
    s = fgets(rdelim, 255, f);
    if (!feof(f) && s == rdelim) { /* Read format */
       rheader[255] = 0;
       s = fgets(rinfo, 255, f);
       if (!feof(f) && s == rinfo) { /* Read infos */
			 rh = rheader;
			 rd = rdelim;
			 ri = rinfo;
			 while (TRUE) {
				 for (i = 0; *rd == '-'; i++) rd++; /* Find columns lenghts */
				 for (i1 = 0; i1 < i; i1++) {
					 s1[i1] = rh[i1];
					 s2[i1] = ri[i1];
             }
				 s1[i1] = s2[i1] = '\0';
				 g_strchug(s1);
				 g_strchomp(s1);
				 g_strdown(s1);
				 if (*s1) *s1 = toupper(*s1);
				 g_strchug(s2);
				 g_strchomp(s2);
				 if (s1[0] && s2[0]
					  && strcmp(s1, "Name") != 0
					  && strcmp(s1, "Description") != 0)
					 l = g_slist_append(l, g_strconcat(s1, " : ", s2, "\n", NULL));
				 rh += i;
				 ri += i;
				 for (i = 0; *rd == ' '; i++) rd++;
				 rh += i;
				 ri += i;
				 if (*rd == '\n' || *rd == '\0') break;
          }
       }
    }
 }
 pclose(f);

 return (l);
}




void gnomame_show_info(GtkToggleButton *b, gamerecord *g, GtkText *info)

{
 gchar *s;
 GSList *i, *i1;

 if (b && b->active && info) {
    gtk_text_freeze(info);
    gtk_text_set_word_wrap(info, TRUE);
    gtk_text_set_point(info, 0);
    gtk_text_forward_delete(info, gtk_text_get_length(info));
    if (g) {
		 for (i = g->errors; i; i = i->next) {
		    if (g->state & GNOMAME_WARNING)
		       s = g_strconcat("WARNING : ", i->data, NULL);
		    else
		       s = g_strconcat("ERROR   : ", i->data, NULL);
          gtk_text_insert(info, errorfont, NULL, NULL, s, -1);
		    g_free(s);
       }
       for (i1 = i = gnomame_readgameinfo(g); i; i = i->next)
          gtk_text_insert(info, infofont, NULL, NULL, i->data, -1);
		 g_slist_foreach(i1, (GFunc) g_free, NULL);
		 g_slist_free(i1);
    }
    gtk_text_thaw(info);
 }
}



void gnomame_show_preview(GtkToggleButton *b, gamerecord *g, GtkWidget *pic)

{
 gchar *s = NULL, *s1;
 FILE *f;

 if (b && b->active && pic) {
	 if (g) {
		 s1 = g_strconcat(gnomamerc->previewdir, "/", g->name, NULL);
		 s = g_strconcat(s1, ".png", NULL);
	         if ((f = fopen(s = g_strconcat(s1, ".png", NULL), "r"))) fclose(f);
		 else if ((f = fopen(s = g_strconcat(s1, ".gif", NULL), "r"))) fclose(f);
		 else if ((f = fopen(s = g_strdup(s1), "r"))) fclose(f);
		 else {
		    g_free(s);
			 s = NULL;
       }
		 g_free(s1);
    }
    if (s == NULL)
		 s = gnome_pixmap_file("/gnomame/gnomame.xpm");
    gnome_pixmap_load_file(GNOME_PIXMAP(pic), s);
	 g_free(s);
 }
}




void gnomame_change_selection(GtkCList *l)

{
 GtkToggleButton *infob, *previewb;

 gnomame_selectedgame = NULL;
 if (l->selection == l->selection_end) {
    if (l->selection)
	    gnomame_selectedgame = gtk_clist_get_row_data(l, GPOINTER_TO_INT(
				                                         l->selection->data));
 } else return; /* Return immediately from multiple selections to prevent
						 a performance loose when screenshot window is active */
 infob = GTK_TOGGLE_BUTTON(lookup_widget(GTK_WIDGET(l), "infob"));
 previewb = GTK_TOGGLE_BUTTON(lookup_widget(GTK_WIDGET(l), "previewb"));
 gnomame_show_info(infob, gnomame_selectedgame, info);
 gnomame_show_preview(previewb, gnomame_selectedgame, preview);
}






void gnomame_show_help(gchar *arg)

{
 gchar *s, *s1;

 if (arg == NULL) arg = "index";
 s1 = g_strconcat(arg, ".html", NULL);
 s = gnome_help_file_find_file("gnomame", s1); g_free(s1);
 if (s == NULL) s = gnome_help_file_find_file("gnomame", "index.html");
 if (s != NULL) {
    s1 = g_strconcat("file:", s, NULL); g_free(s);
    gnome_url_show(s1); g_free(s1);
 } else
	 gnomame_error("No help installed.", GNOMAME_WARNING);
}



static void gnomame_mark_game(GtkCList *notepad, gamerecord *g)

{
 gchar *row[2];
 gamerecord *gg;
 gint i;
 gint cmp;
 

 if (g) {
    for (i = 0; i < notepad->rows; i++) {
	    gg = gtk_clist_get_row_data(notepad, i);
		 cmp = strcmp(gg->name, g->name);
		 if (cmp == 0) return;
		 if (cmp > 0) break;
    }
	 g->marked = TRUE;
    row[0] = g->name;
    row[1] = g->description;
    gtk_clist_insert(notepad, i, row);
	 gtk_clist_set_row_data(notepad, i, g);
 }
}




void gnomame_mark_selected(GtkCList *notepad)

{
 gamerecord *g;
 GList *l1;

 gtk_clist_freeze(listw);
 gtk_clist_freeze(notepad);
 for (l1 = listw->selection; l1; l1 = l1->next) {
	 g = gtk_clist_get_row_data(listw, GPOINTER_TO_INT(l1->data));
	 gnomame_mark_game(notepad, g);
	 gnomame_refresh_row(GPOINTER_TO_INT(l1->data));
 }
 gtk_clist_thaw(notepad);
 gtk_clist_thaw(listw);
 notepad_updatestatus();
}




void gnomame_unmark_selected(GtkCList *notepad)

{
 gamerecord *g;
 GList *l1;
 gint i;

 gtk_clist_freeze(listw);
 gtk_clist_freeze(notepad);
 for (l1 = listw->selection; l1; l1 = l1->next) {
	 g = gtk_clist_get_row_data(listw, GPOINTER_TO_INT(l1->data));
    if ((i = gtk_clist_find_row_from_data(notepad, g)) >= 0) {
	    gtk_clist_remove(notepad, i);
	    g->marked = FALSE;
	    gnomame_refresh_row(GPOINTER_TO_INT(l1->data));
    }
 }
 gnomame_unselect_all();
 gtk_clist_thaw(notepad);
 gtk_clist_thaw(listw);
 notepad_updatestatus();
}




static GSList *gnomame_givemarked(GtkCList *notep)

{
 GSList *l = NULL;
 gamerecord *g;
 int i;

 for (i = 0; i < notep->rows; i++) {
	 g = (gamerecord *) gtk_clist_get_row_data(notep, i);
	 l = g_slist_append(l, g_strdup(g->name));
 }

 return (l);
}




static void gnomame_marklist(GtkCList *notepad, GSList *l)

{
 gamerecord *g;
 gchar *row[2];
 int i;

 gtk_clist_freeze(listw);
 gtk_clist_freeze(notepad);
 while (l) {
	 for (i = 0; i < ngames; i++) {
		 g = (gamerecord *) games[i];
       if (strcmp(g->name, l->data) == 0) {
	       g->marked = TRUE;
          row[0] = g->name;
          row[1] = g->description;
          gtk_clist_append(notepad, row);
	       gtk_clist_set_row_data(notepad, notepad->rows - 1, g);
			 gnomame_refresh_game(g);
       }
    }
	 l = l->next;
 }
 gtk_clist_thaw(notepad);
 gtk_clist_thaw(listw);
}




void gnomame_select_all(void)

{
 gtk_clist_select_all(listw);
}



void gnomame_unselect_all(void)

{
 gtk_clist_unselect_all(listw);
}




void gnomame_goto_game(gchar *name)

{
 gint i;
 gamerecord *g;

 if (name && *name != 0) {
    for (i = 0; i < listw->rows; i++) {
	    g = gtk_clist_get_row_data(listw, i);
	    if (strcmp(name, g->name) == 0) {
		    gnomame_unselect_all();
		    gtk_clist_moveto(listw, i, 0, 0.5, 0);
		    gtk_clist_select_row(listw, i, 0);
		    break;
       }
    }
 }
}





int gnomame_plugin_code_to_num(GSList *l, gchar *code)

{
 int i = 0;

 while (l) {
	 if (strcmp(((xmame_plugin *) l->data)->code, code) == 0) break;
    l = l->next;
	 i++;
 }
 return (i);
}




static GSList *gnomame_get_plugins(gchar *plugtype)

{
 xmame_plugin *plug;
 GSList *l = NULL;
 gchar *s, *s1, riga[256], code[256], description[256];
 FILE *f;

 s = g_strconcat("-list-", plugtype, "-plugins", NULL);
 f = gnomame_pipe(s, "");
 g_free(s);
 s = fgets(riga, 255, f); if (s != riga) return (l);
 s = fgets(riga, 255, f); if (s != riga) return (l);
 while (!feof(f)) {
    s = fgets(riga, 255, f); if (s != riga) break;
	 while (*s == ' ' || *s == '\t') s++;
	 for (s1 = code; isalnum(*s); s++, s1++) *s1 = *s;
	 *s1 = '\0';
	 while (*s == ' ' || *s == '\t') s++;
	 for (s1 = description; *s != '\0' && *s != '\n'; s++, s1++) *s1 = *s;
	 *s1 = '\0';
	 if ((plug = g_malloc(sizeof(xmame_plugin))) == NULL)
		 gnomame_error("No memory for plugins list", GNOMAME_FATAL);
	 plug->code = g_strdup(code);
    g_strchomp(description);
	 plug->description = g_strdup(description);
	 l = g_slist_append(l, plug);
 }
 pclose(f);
 return (l);
}





gchar *gnomame_plugin_code(GSList *l, int i)

{
 while (l && i--) l = l->next;
 return ((xmame_plugin *) l->data)->code;
}




GList *gnomame_get_plugin_descriptions(GSList *l)

{
 GList *l1 = NULL;

 while (l) {
	 l1 = g_list_append(l1, ((xmame_plugin *) l->data)->description);
	 l = l->next;
 }
 return (l1);
}




static void free_plugin(xmame_plugin *plug)

{
 g_free(plug->code);
 g_free(plug->description);
 g_free(plug);
}




static void pixmap_to_title(GtkWidget *mainw, GtkCList *listw, gchar *pixname, int col)

{
 gchar *filename;
 GtkWidget *pixmap;

 pixmap = gtk_type_new (gnome_pixmap_get_type ());
 filename = gnome_pixmap_file(pixname);
 if (filename)
    gnome_pixmap_load_file (GNOME_PIXMAP (pixmap), filename);
 else
    g_warning (_("Couldn't find pixmap file: %s"), pixname);
 g_free (filename);
 gtk_widget_ref (pixmap);
 gtk_object_set_data_full(GTK_OBJECT(mainw), listw_col[col].code, pixmap,
                         (GtkDestroyNotify) gtk_widget_unref);
 gtk_widget_show (pixmap);
 gtk_clist_set_column_widget(listw, col, pixmap);
}




static void col_add(int i, gchar *n, gchar *c, gint w)

{
 listw_col[i].name = n;
 listw_col[i].code = c;
 listw_col[i].width = w;
}






static GtkCList *create_listw(GtkWidget *mainw)
	
{
 GtkWidget *listw;
 gint i;

 listw_colnum = 0;
 col_add(listw_colnum++, NULL, "0", 20);
 col_add(listw_colnum++, NULL, "1", 20);
 col_add(listw_colnum++, NULL, "2", 20);
 col_add(listw_colnum++, "Description", "DESC", 290);
 if (gnomamerc->showcol[0]) col_add(listw_colnum++, "Name", "NAME", 80);
 if (gnomamerc->showcol[1]) col_add(listw_colnum++, "Year", "YEAR", 80);
 if (gnomamerc->showcol[2]) col_add(listw_colnum++, "Manufacturer", "MANU", 80);
 if (gnomamerc->showcol[3]) col_add(listw_colnum++, "Clone", "CLON", 80);
 listw = gtk_clist_new (listw_colnum);
 gtk_widget_ref (listw);
 gtk_object_set_data_full(GTK_OBJECT(mainw), "listw", listw,
                            (GtkDestroyNotify) gtk_widget_unref);
 gtk_container_add(GTK_CONTAINER(lookup_widget(mainw, "scrollw")), listw);
 gtk_clist_set_selection_mode (GTK_CLIST (listw), GTK_SELECTION_EXTENDED);
 gtk_clist_column_titles_show (GTK_CLIST (listw));

 pixmap_to_title(mainw, GTK_CLIST(listw), "gnomame/greenv.xpm", 0);
 pixmap_to_title(mainw, GTK_CLIST(listw), "gnomame/rcfile.xpm", 1);
 pixmap_to_title(mainw, GTK_CLIST(listw), "gnomame/marker.xpm", 2);

 for (i = 0; i < listw_colnum; i++) {
    gtk_clist_set_column_width(GTK_CLIST(listw), i, listw_col[i].width);
    if (listw_col[i].name)
       gtk_clist_set_column_widget(GTK_CLIST(listw), i, gtk_label_new(_(listw_col[i].name)));
 }
 set_first_column(GTK_CLIST(listw), filter.state);

 gtk_signal_connect (GTK_OBJECT (listw), "click_column",
                     GTK_SIGNAL_FUNC (on_listw_click_column), NULL);
 gtk_signal_connect (GTK_OBJECT (listw), "select_row",
                     GTK_SIGNAL_FUNC (on_listw_select_row), NULL);
 gtk_signal_connect (GTK_OBJECT (listw), "unselect_row",
                     GTK_SIGNAL_FUNC (on_listw_unselect_row), NULL);
 gtk_signal_connect (GTK_OBJECT (listw), "key_press_event",
                     GTK_SIGNAL_FUNC (on_listw_key_press_event), NULL);
 gtk_signal_connect (GTK_OBJECT (listw), "button_press_event",
                     GTK_SIGNAL_FUNC (on_listw_button_press_event), NULL);
 gtk_widget_show(listw);

 return (GTK_CLIST(listw));
}




void gnomame_query_xmame(GtkWidget *mainw)

{
 xmameprefrule *r;

 if (listw) gtk_widget_destroy(GTK_WIDGET(listw));
 listw = create_listw(mainw);
 if ((xmameversion = gnomame_xmame_version()) == NULL)
	 gnomame_error("Could not execute xmame", GNOMAME_WARNING);
 else {
    gnomame_set_status("Loading xmame options...");
    xmameprefs_free(xmamerc);
    xmameprefs_loadrules();
    xmamerc = xmameprefs_read(NULL);
    if ((r = xmameprefrule_find("sound-mixer-plugin")) && r->valid) {
       if (gnomame_mixerplugins) {
	       g_slist_foreach(gnomame_mixerplugins, (GFunc) free_plugin, NULL);
	       g_slist_free(gnomame_mixerplugins);
       }
       gnomame_mixerplugins = gnomame_get_plugins("mixer");
    }
    if ((r = xmameprefrule_find("dsp-plugin")) && r->valid) {
       if (gnomame_dspplugins) {
	       g_slist_foreach(gnomame_dspplugins, (GFunc) free_plugin, NULL);
	       g_slist_free(gnomame_dspplugins);
       }
       gnomame_dspplugins = gnomame_get_plugins("dsp");
    }
    gnomame_load_games();
    gnomame_marklist(GTK_CLIST(lookup_widget(notepad, "notelist")), gnomamerc->marked);
 }
}




void gnomame_init(GtkWidget *mainw)


{
 memset(&filter, 0, sizeof(filter));
 filter.state = GNOMAME_OK;
 if (gnomamerc->status > 0) filter.state |= GNOMAME_WARNING;
 if (gnomamerc->status > 1) filter.state |= GNOMAME_FATAL;
 if (gnomamerc->status > 2) filter.state |= GNOMAME_NOROM;
 availpix.pix = noavailpix.pix = qmarkpix.pix = NULL;
 load_pixmaps(mainw);
/* infofont = gdk_font_load("-adobe-courier-medium-r-*-*-*-140-*-*-*-*-*-*"); */
 infofont = NULL;  /* Use default font */
 errorfont = gdk_font_load("-misc-fixed-bold-r-*-*-*-*-*-*-*-*-*-*");
 gnomame_query_xmame(mainw);
}




void gnomame_exit(void)

{
 GSList *l;

 if (gnomamerc->marked) {
	 g_slist_foreach(gnomamerc->marked, (GFunc) g_free, NULL);
	 g_slist_free(gnomamerc->marked);
 }
 gnomamerc->marked = gnomame_givemarked(GTK_CLIST(lookup_widget(notepad,
				                                                    "notelist")));
 gnomameprefs_write(gnomamerc);
 gnomameprefs_free(gnomamerc);
 for (l = gnomame_mixerplugins; l; l = l->next) {
	 g_free(((xmame_plugin *) l->data)->code);
	 g_free(((xmame_plugin *) l->data)->description);
 }
 g_slist_foreach(gnomame_mixerplugins, (GFunc) g_free, NULL);
 g_slist_free(gnomame_mixerplugins);
 for (l = gnomame_dspplugins; l; l = l->next) {
	 g_free(((xmame_plugin *) l->data)->code);
	 g_free(((xmame_plugin *) l->data)->description);
 }
 g_slist_foreach(gnomame_dspplugins, (GFunc) g_free, NULL);
 g_slist_free(gnomame_dspplugins);
 gameslist_free();
 xmameprefs_free(xmamerc);
 gdk_pixmap_unref(availpix.pix);
 gdk_pixmap_unref(noavailpix.pix);
 gdk_pixmap_unref(qmarkpix.pix);
 gdk_pixmap_unref(redqmarkpix.pix);
 gdk_pixmap_unref(rcfile.pix);
 gdk_pixmap_unref(marker.pix);
}



static gboolean is_directory(const gchar *directory, const gchar *file) {
	gchar *s;
	struct stat file_info;
	
	s = g_strconcat(directory, "/", file, NULL);	

	if(!stat(s, &file_info)) {
		g_free(s);
		/* got stat info */
		if(S_ISDIR(file_info.st_mode))
			return TRUE;
		else
			return FALSE;
	} else {
		g_free(s);
		/* stat info failed */
		return FALSE;
	}
}
