#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>
#include <locale.h>
#include <gnome.h>
#include "preferences.h"
#include "core.h"





static xmameprefrule *rules = NULL;
static int nrules = 0;





xmameprefrule *xmameprefrule_get(int i)

{
 return (i < nrules) ? (rules + i) : NULL;
}





xmameprefrule *xmameprefrule_find(gchar *name)

{
 int i;

 for (i = 0; i < nrules; i++)
	 if (strcmp(rules[i].name, name) == 0) return (rules + i);

 return NULL;
}




static void xmamepref_del(xmamepref *p)

{
 switch (p->r->type) {
 case XMAMEPREF_TYPE_STRING :
	 if (p->val.s) g_free(p->val.s);
	 break;
 case XMAMEPREF_TYPE_LIST :
	 g_slist_foreach(p->val.l, (GFunc) g_free, NULL);
	 g_slist_free(p->val.l);
	 break;
 }
 g_free(p);
}




void xmameprefs_del(xmameprefs *prefs, xmamepref *p)

{
 if (p) {
    prefs->prefs = g_slist_remove(prefs->prefs, p);
    xmamepref_del(p);
 }
}




xmamepref *xmameprefs_add(xmameprefs *prefs, xmameprefrule *rule)

{
 xmamepref *p;

 if ((p = g_malloc(sizeof(xmamepref))) == NULL)
	 gnomame_error("No memory for preferences allocation.", GNOMAME_FATAL);
 p->r = rule;
 switch (rule->type) {
 case XMAMEPREF_TYPE_BOOL : 
 case XMAMEPREF_TYPE_FLAG : 
    p->val.b = FALSE;
 case XMAMEPREF_TYPE_INT : 
	 p->val.i = 0;
	 break;
 case XMAMEPREF_TYPE_FLOAT : 
	 p->val.f = 0;
	 break;
 case XMAMEPREF_TYPE_NONE : 
 case XMAMEPREF_TYPE_STRING : 
	 p->val.s = NULL;
	 break;
 case XMAMEPREF_TYPE_LIST : 
	 p->val.l = NULL;
	 break;
 default :
	 gnomame_error("Trying to create a bad type option", GNOMAME_FATAL);
	 break;
 }
 prefs->prefs = g_slist_append(prefs->prefs, p);

 return (p);
}




xmamepref *xmameprefs_find(xmameprefs *prefs, gchar *name)

{
 xmamepref *p;
 xmameprefrule *r;
 GSList *l = prefs->prefs;

 while (l != NULL) {
	 p = (xmamepref *) l->data;
	 if (strcmp(p->r->name, name) == 0) return (p);
	 l = l->next;
 }
 r = xmameprefrule_find(name);
 if (r) return xmameprefs_add(prefs, r);

 return (NULL);
}




static void write_double(FILE *f, gchar *label, gdouble d)

{
 if (d - (double) ((int) d) != 0)
    fprintf(f, "%s\t%f\n", label, d);
 else
    fprintf(f, "%s\t%d\n", label, (int) d);
}



static void write_string(FILE *f, gchar *label, gchar *s)

{
 if (s != NULL && s[0] != '\0') fprintf(f, "%s\t%s\n", label, s);
}




static void write_int(FILE *f, gchar *label, gint i)

{
 fprintf(f, "%s\t%d\n", label, i);
}




static void write_bool(FILE *f, gchar *label, gboolean b)

{
 fprintf(f, "%s\t%d\n", label, b ? 1 : 0);
}




static void write_flag(FILE *f, gchar *label, gboolean b)

{
 if (b) fprintf(f, "%s\n", label);
}




static void write_list(FILE *f, gchar *label, GSList *l)

{
 if (l) {
    if (strcmp(label, "rompath") == 0) {
       fprintf(f, "%s %s", label, (char *) l->data);
       for (l = l->next; l; l = l->next) fprintf(f, ":%s", (char *) l->data);
       fprintf(f, "\n");
    } else
       for (; l; l = l->next) fprintf(f, "%s %s\n", label, (char *) l->data);
 }
}




static gboolean check_valid(gchar *name)

{
 xmameprefrule *r;
 char s[256];

 if (name) {
    r = xmameprefrule_find(name);
	 if (r == NULL) {
		 sprintf(s, "No rule for option %s\n, check configuration file", name);
		 gnomame_error(s, GNOMAME_WARNING);
		 return (FALSE);
    } else
       return (r->valid && check_valid(r->depend));
 } else
	 return (TRUE);
}




void xmameprefs_loadrules(void)

{
 FILE *f;
 char *s, riga[256], *filename;
 char varname[256], otype, wtype, depend[256], cond[256];
 xmameprefrule *r;
 GSList *l = NULL, *l1 = NULL;
 int i, oldnrules;

 oldnrules = nrules;
 nrules = 0;
 filename = gnome_libdir_file("gnomame/etc/xmameoptions");
 if ((f = fopen(filename, "r"))) {
	 riga[255] = '\0';
	 while (!feof(f)) {
		 s = fgets(riga, 255, f);
		 if (s != riga) break;
		 while (*s == ' ' || *s == '\t') s++;
		 if (*s && *s != '#' && *s != '\n') {
		    i = sscanf(s, "%s %c %c %s %s",
					      varname, &otype, &wtype, depend, cond);
		    if (i >= 3 && varname[0]) {
				 if ((r = malloc(sizeof(xmameprefrule))) != NULL) {
					 r->name = g_strdup(varname);
					 r->type = otype;
					 r->widgtype = wtype;
					 r->depend = NULL;
					 if (i >= 4) {
					    r->depend = g_strdup(depend);
						 r->negative = TRUE;
						 r->condvalue = 0;
						 if (i >= 5) {
							 r->negative = cond[0] == '!';
							 if (r->negative)
								 r->condvalue = atoi(cond + 1);
                      else
								 r->condvalue = atoi(cond);
                   }
                }
					 r->valid = gnomame_check_option(varname);
				    l = g_slist_append(l, r);
					 nrules++;
             } else
					 gnomame_error("No room to alloc memory for preference's rules",
							         GNOMAME_FATAL);
          } else if (i >= 1)
				 gnomame_error("Options config file corrupted", GNOMAME_FATAL);
       }
    }
	 fclose(f);
 }
 g_free(filename);

 /* Put the rules GSList into the array rules */

 if (rules) {
	 for (i = 0; i < oldnrules; i++) {
		 g_free(rules[i].name);
		 g_free(rules[i].depend);
    }
	 g_free(rules);
 }
 rules = g_malloc(sizeof(xmameprefrule) * nrules);
 if (rules == NULL)
	 gnomame_error("No room to allocate the rules array", GNOMAME_FATAL);
 for (l1 = l, r = rules; l1; l1 = l1->next, r++)
	 *r = *((xmameprefrule *) l1->data);
 g_slist_foreach(l, (GFunc) g_free, NULL);
 g_slist_free(l);
 for (i = 0; i < nrules; i++)
	 if (rules[i].valid) rules[i].valid = check_valid(rules[i].depend);
}



void xmameprefs_free(xmameprefs *p)

{
 if (p) {
    g_free(p->rcfilename);
    g_slist_foreach(p->prefs, (GFunc) xmamepref_del, NULL);
    g_slist_free(p->prefs);
 }
}




static gboolean check_dir(gchar *name)
	
{
 DIR *d;
 gchar *s, s1[256];
 gint i;

 i = strlen(name) - 1;
 while (i >= 0 && name[i] != '/') i--;
 if (i < 0) return (TRUE);
 for (s = s1; i > 0; i--) *s++ = *name++;
 *s = '\0';
 d = opendir(s1);
 if (d == NULL) {
	 if (mkdir(s1, S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH) == 0)
		 return (TRUE);
    else {
		 s = g_strconcat("Could not create directory : ", s1, NULL);
	    gnomame_error(s, GNOMAME_WARNING);
		 g_free(s);
	    return (FALSE);
    }
 } else {
	 closedir(d);
    return (TRUE);
 }
}




void xmameprefs_write(xmameprefs *prefs)

{
 FILE *f;
 GSList *l;
 gchar s[256], *s1;
 gchar *loca;
 xmamepref *p;

 loca = g_strdup(setlocale(LC_ALL, NULL));
 setlocale(LC_ALL, "C");        /* For writing of point in float values */
 check_dir(s1 = g_strconcat(g_get_home_dir(), "/.xmame/", NULL));
 g_free(s1);
 if (prefs->rcfilename == NULL) {
	 /* NULL rcfilename : try to write a ~/.xmame/xmamerc file */
	 prefs->rcfilename = g_strconcat(g_get_home_dir(), "/.xmame/xmamerc", NULL);
 }
 check_dir(prefs->rcfilename);
 if ((f = fopen(prefs->rcfilename, "w")) != NULL) {
	 p = xmameprefs_find(prefs, "widthscale");
	 if (p->val.f == xmameprefs_find(prefs, "heightscale")->val.f) {
		 xmameprefs_find(prefs, "scale")->val.f = p->val.f;
		 xmameprefs_del(prefs, p);
       p = NULL;
		 xmameprefs_del(prefs, xmameprefs_find(prefs, "heightscale"));
    }
    if (xmameprefs_find(prefs, "autodouble")->val.b)
       xmameprefs_del(prefs, p);
		 xmameprefs_del(prefs, xmameprefs_find(prefs, "heightscale"));
		 xmameprefs_del(prefs, xmameprefs_find(prefs, "scale"));
	 for (l = prefs->prefs; l; l = l->next) {
		 p = (xmamepref *) l->data;
		 if (p->r->valid) {
		    switch (p->r->type) {
          case XMAMEPREF_TYPE_BOOL :
	          write_bool(f, p->r->name, p->val.b);
			    break;
          case XMAMEPREF_TYPE_FLAG :
	          write_flag(f, p->r->name, p->val.b);
			    break;
          case XMAMEPREF_TYPE_INT :
	          write_int(f, p->r->name, p->val.i);
			    break;
          case XMAMEPREF_TYPE_FLOAT :
	          write_double(f, p->r->name, p->val.f);
			    break;
          case XMAMEPREF_TYPE_STRING :
	          write_string(f, p->r->name, p->val.s);
			    break;
          case XMAMEPREF_TYPE_LIST :
	          write_list(f, p->r->name, p->val.l);
			    break;
          }
       }
    }
    fclose(f);
	 if (prefs->g) prefs->g->haverc = TRUE;
 } else {
    sprintf(s, "Error writing rc file %s", prefs->rcfilename);
	 gnomame_error(s, GNOMAME_WARNING);
 }
 setlocale(LC_ALL, loca);
 g_free(loca);
}




static GSList *give_paths(gchar *romp)

{
 gchar s[1024];
 GSList *l = NULL;
 int i;

 if (romp != NULL) {
    while (*romp != '\0') {
		 while (*romp == ':') romp++;
       for (i = 0; *romp != '\0' && *romp != ':'; i++) s[i] = *romp++;
       s[i] = '\0';
		 if (g_slist_find_custom(l, s, (GCompareFunc) strcmp) == NULL)
          l = g_slist_append(l, g_strdup(s));
    }
 }
 return (l);
}



xmameprefs *xmameprefs_read(gamerecord *g)

{
 FILE *f = NULL;
 char s[256], s1[256], s2[256];
 char *tmp, *tmp1;
 gchar *rc = NULL;
 xmameprefs *prefs;
 xmamepref *p;

 if ((prefs = (xmameprefs *) malloc(sizeof(xmameprefs))) == NULL)
	 gnomame_error("Could not allocate memory for preferences", GNOMAME_FATAL);
 prefs->g = NULL;
 prefs->rcfilename = NULL;
 prefs->prefs = NULL;
 if (rules == NULL) xmameprefs_loadrules();
 if (g) {
	 rc = g_strconcat(g_get_home_dir(), "/.xmame/rc/", g->name, "rc", NULL);
    f = fopen(rc, "r");
/*	 if (f == NULL) {
		 sprintf(s, "Cannot open file %s", rc);
		 gnomame_error(s, GNOMAME_WARNING);
    } */
 }
 if (f == NULL) {
    tmp = g_strconcat(g_get_home_dir(), "/.xmame/xmamerc", NULL);
    f = fopen(tmp, "r");
	 if (g == NULL) rc = tmp; else g_free(tmp);
 }
 if (f == NULL) {
    tmp = g_strconcat(gnomamerc->xmamepath, " -showconfig 2>/dev/null", NULL);
    f = popen(tmp, "r");
    g_free(tmp);
    if (f == NULL) {
		 /* Could not read any xmame configuration so return with an empty
		  * configuration structure and show a warning message */
		 gnomame_error("Could not read any xmame configuration", GNOMAME_WARNING);
    }
 }
 if (f) {
	 prefs->rcfilename = rc;
	 prefs->g = g;
    while (!feof(f)) {
       tmp = fgets(s, 256, f);
		 if (tmp != s) break;
       g_strchug(s);
       if (s[0] != '#') {
          for (tmp = s, tmp1 = s1; *tmp > ' '; tmp++) *tmp1++ = *tmp;
          *tmp1 = '\0';
          for (;*tmp != '\0' && *tmp <= ' '; tmp++) ;
          for (tmp1 = s2; *tmp > ' '; tmp++) *tmp1++ = *tmp;
          *tmp1 = '\0';
          if (s1[0] != '\0') {
				 p = xmameprefs_find(prefs, s1);
				 if (p != NULL) {
         	    switch (p->r->type) {
                case XMAMEPREF_TYPE_BOOL :
                   p->val.b = !(strcmp(s2, "0") == 0);
      			    break;
                case XMAMEPREF_TYPE_FLAG :
                   p->val.b = TRUE;
      			    break;
                case XMAMEPREF_TYPE_INT :
                   p->val.i = atoi(s2);
   	   		    break;
                case XMAMEPREF_TYPE_FLOAT :
                   p->val.f = g_strtod(s2, NULL);
      			    break;
                case XMAMEPREF_TYPE_STRING :
                   if (p->val.s) g_free(s);
   					 p->val.s = g_strdup(s2);
        				 break;
                case XMAMEPREF_TYPE_LIST :
   				    if (strcmp(s1, "rompath") == 0) {
   					    g_slist_foreach(p->val.l, (GFunc) g_free, NULL);
   					    g_slist_free(p->val.l);
                      p->val.l = give_paths(s2);
                   } else
                      p->val.l = g_slist_append(p->val.l, g_strdup(s2));
      			    break;
                }
				 }
          }
       }
    }
    pclose(f);
	 if ((p = xmameprefs_find(prefs, "scale"))->val.f != 0) {
		 xmameprefs_find(prefs, "widthscale")->val.f = p->val.f;
		 xmameprefs_find(prefs, "heightscale")->val.f = p->val.f;
    }
 }
 return (prefs);
}



void xmameprefs_delrc(gamerecord *g)

{
 gchar *rc;
 FILE *f;
 gchar s[256];

 rc = g_strconcat(g_get_home_dir(), "/.xmame/rc/", g->name, "rc", NULL);
 if ((f = fopen(rc, "r")) != NULL) {
	 fclose(f);
	 if (unlink(rc) != 0) {
		 sprintf(s, "Cannot delete the file : %s", rc);
		 gnomame_error(s, GNOMAME_WARNING);
    } else
       g->haverc = FALSE;
 }
 g_free(rc);
}



gnomameprefs *gnomameprefs_read(void)

{
 gnomameprefs *prefs;
 gchar *s;
 FILE *f;
 char riga[256], s1[256], s2[256];
 gint i;

 if ((prefs = g_malloc0(sizeof(gnomameprefs)))) {
    s = g_strconcat(g_get_home_dir(), "/.gnome/gnomame", NULL);
    if ((f = fopen(s, "r")) != NULL) {
	    while (!feof(f)) {
			 s = fgets(riga, 256, f);
			 if (s != riga) break;
	       sscanf(s, " %s %s", s1, s2);
          if (strcmp(s1, "showinfo") == 0)
			    prefs->showinfo = (s2[0] != '0');
		    else if (strcmp(s1, "status") == 0)
			    prefs->status = atoi(s2);
	       else if (strcmp(s1, "showpreview") == 0)
			    prefs->showpreview = (s2[0] != '0');
	       else if (strcmp(s1, "confirmplay") == 0)
			    prefs->confirmplay = (s2[0] != '0');
	       else if (strcmp(s1, "toolbartype") == 0)
			    prefs->toolbartype = atoi(s2);
	       else if (strcmp(s1, "xmamepath") == 0)
			    prefs->xmamepath = g_strdup(s2);
		    else if (strcmp(s1, "previewdir") == 0)
			    prefs->previewdir = g_strdup(s2);
		    else if (strncmp(s1, "showcol_", 8) == 0) {
				 i = atoi(s1 + 8);
				 if (i >= 0 && i < 4)
			       prefs->showcol[i] = (s2[0] != '0');
		    } else if (strcmp(s1, "marked") == 0) {
	          while (!feof(f)) {
	             s = fgets(s1, 10, f);
					 if (s != s1) break;
					 while (*s != '\n' && *s != '\0') s++;
					 *s = '\0';
					 prefs->marked = g_slist_append(prefs->marked, g_strdup(s1));
             }
          }
	    }
	    fclose(f);
    }
    g_free(s);
    if (prefs->xmamepath == NULL)
		 prefs->xmamepath = g_strdup("xmame");
    if (prefs->previewdir == NULL)
		 prefs->previewdir = g_strdup("/usr/games/xmame/images");
 }

 return (prefs);
}



void gnomameprefs_write(gnomameprefs *prefs)

{
 char s[256];
 gchar *filename;
 FILE *f;
 GSList *l;
 gint i;

 filename = g_strconcat(g_get_home_dir(), "/.gnome/gnomame", NULL);
 if ((f = fopen(filename, "w")) != NULL) {
    write_string(f, "xmamepath", prefs->xmamepath);
    write_string(f, "previewdir", prefs->previewdir);
    write_bool(f,   "showpreview", prefs->showpreview);
    write_bool(f,   "showinfo", prefs->showinfo);
    write_bool(f,   "confirmplay", prefs->confirmplay);
    write_int(f,    "toolbartype", prefs->toolbartype);
    write_int(f,    "status", prefs->status);
	 for (i = 0; i < 4; i++) {
		 sprintf(s, "showcol_%d", i);
		 write_bool(f, s, prefs->showcol[i]);
    }
	 if (prefs->marked) {
		 fprintf(f, "marked\n");
		 for (l = prefs->marked; l; l = l->next)
		    fprintf(f, "%s\n", (gchar *) l->data);
	 }
	 fclose(f);
 } else {
	 sprintf(s, "Error writing preferences file : %s\n", filename);
	 gnomame_error(s, GNOMAME_WARNING);
 }
 g_free(filename);
}




void gnomameprefs_free(gnomameprefs *p)

{
 if (p) {
    g_free(p->xmamepath);
    g_free(p->previewdir);
	 g_slist_foreach(p->marked, (GFunc) g_free, NULL);
	 g_slist_free(p->marked);
    g_free(p);
 }
}
